;;; Assembler source for FPA support code and emulator - PIE veneer
;;; Post-core part.
; Copyright (C) Advanced RISC Machines Limited, 1992-7 (now named ARM Limited).
; All rights reserved.
;
; Redistribution and use in source and binary forms, with or without
; modification, are permitted provided that the following conditions are met:
;
;     * Redistributions of source code must retain the above copyright notice,
;       this list of conditions and the following disclaimer.
;     * Redistributions in binary form must reproduce the above copyright
;       notice, this list of conditions and the following disclaimer in the
;       documentation and/or other materials provided with the distribution.
;     * Neither the name of Advanced RISC Machines Limited nor ARM Limited nor
;       the names of its contributors may be used to endorse or promote
;       products derived from this software without specific prior written
;       permission.
;
; THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
; AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
; IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
; ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
; LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
; CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
; SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
; INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
; CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
; ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
; POSSIBILITY OF SUCH DAMAGE.

; The initialisation routine
; ==========================

        GBLS    InstallName
InstallName     SETS         "FP":CC:"$FPCodeType":CC:"_DeInstall"
        EXPORT          $InstallName
$InstallName
        B       core_shutdown

        GBLS    DeInstallName
DeInstallName   SETS       "FP":CC:"$FPCodeType":CC:"_Install"
        EXPORT          $DeInstallName
$DeInstallName
        STMFD   r13!, {r7-r11, lr}

        ADRL    r10, FP_Workspace

; First thing we need to do is determine where the old undefined instruction
; handler was. We do this by decoding the instruction at the undefined
; instruction vector. (Note we're assuming that we're using the "StandAlone"
; undefined instruction vector linkage here.)

        MOV     r1, #Undef_vector
        LDR     r1, [r1]
        STR     r1, OldVector
        AND     r0, r1, #&ff000000
        TEQ     r0, #BranchConvertOpcode
        BNE     Module_Init_NotBranchAt4
        MOV     r1, r1, LSL #2
        ADD     r1, r1, #BranchConvertOffset
        BIC     r1, r1, #&fc000000

        STR     r1, BranchTarget
        ADR     r1, BranchTarget

        B       Module_Init_GotOldHandler

Module_Init_NotBranchAt4
        TEQ     r0, #&e5000000               ;Check for LDR pc, [pc, #xxx]
        ANDEQ   r0, r1, #&ff0000
        TEQEQ   r0, #&9f0000
        ANDEQ   r0, r1, #&f000
        TEQEQ   r0, #&f000
        BNE     Module_Init_BadVector
        MOV     r1, r1, LSL #20
        MOV     r1, r1, LSR #20
        ADD     r1, r1, #BranchConvertOffset

Module_Init_GotOldHandler

; Now initialise the core. Note that we're assuming MultipleContexts =
; {FALSE} here.

        BL      core_initws

; Can we proceed? The answer is yes if we have found no hardware and the FPE
; code is present, or if we found the FPA hardware.

        CMP     r0, #SysID_FPA
        CMPNE   r0, #-1
        BNE     Module_Init_BadHardware

; Now we're ready to activate the (one and only) floating point context.

        BL      core_activatecontext

Module_Init_Success
        CMP     r0, #0                      ;Clears V
        LDMFD   r13!, {r7-r11, pc}

Module_Init_BadVector
        ADR     r0, Error_BadVector
        CMP     r0, #&80000000               ;Sets V
        LDMFD   r13!, {r7-r11, pc}

Error_BadVector
        &       &ffffff
        =       "Initialisation failure: strange previous undef. vector", 0
        ALIGN

Module_Init_BadHardware
        ADR     r0, Error_BadHardware
        CMP     r0, #&80000000               ;Sets V
        LDMFD   r13!, {r7-r11, pc}

Error_BadHardware
        &       &ffffff
        =       "Initialisation failure: unknown hardware or no hardware found", 0
        ALIGN

; The core/veneer interface
; =========================

; Routine to install a new undefined instruction handler. Note this is
; definitely 26-bit code or possibly 32-bit code with suitable restrictions
; on the memory map: a 32-bit system can't enter an arbitrarily placed
; routine directly via a branch from the page 0 vectors.
;
; Entry: r1 = new undefined instruction handler (since
;             UndefHandlerConvention = "StandAlone");
;        r13 is the stack pointer;
;        r14 is the return link.
;
; Exit:  All registers preserved.

veneer_newhandlers

        STMFD   r13!, {r0, lr}
        SUB     r0, r1, #BranchConvertOffset
        BIC     r0, r0, #&fc000000
        MOV     r0, r0, LSR #2
        ADD     r0, r0, #BranchConvertOpcode
        MOV     lr, #Undef_vector
        STR     r0, [lr]
        LDMFD   r13!, {r0, pc}

; Routine to restore the original undefined instruction handler. Note this
; is definitely 26-bit code or possibly 32-bit code with suitable
; restrictions on the memory map: a 32-bit system can't enter an arbitrarily
; placed routine directly via a branch from the page 0 vectors.
;
; Entry: r13 is the stack pointer;
;        r14 is the return link.
;
; Exit:  All registers preserved.

veneer_restorehandlers

        STMFD   r13!, {r0, lr}
        LDR     r0, OldVector
        MOV     lr, #Undef_vector
        STR     r0, [lr]
        LDMFD   r13!, {r0, pc}

; Trap handler routines
; =====================

veneer_invalidop1_single
veneer_invalidop2_single
veneer_invalidop1_double
veneer_invalidop2_double
veneer_invalidop1_extended
veneer_invalidop2_extended
veneer_invalidop1_integer
veneer_invalidop2_integer
veneer_invalidop1_packed
veneer_invalidop1_xpacked
veneer_invalidop1_ldfp
veneer_invalidop1_ldfpx

        ADR     r0, E_FP_IVO
        B       GenerateError

veneer_zerodivide1_single
veneer_zerodivide2_single
veneer_zerodivide1_double
veneer_zerodivide2_double
veneer_zerodivide1_extended
veneer_zerodivide2_extended

        ADR     r0, E_FP_DVZ
        B       GenerateError

veneer_overflow_single
veneer_overflow_double
veneer_overflow_extended

        ADR     r0, E_FP_OFL
        B       GenerateError

veneer_underflow_single
veneer_underflow_double
veneer_underflow_extended

        ADR     r0, E_FP_UFL
        B       GenerateError

veneer_inexact_single
veneer_inexact_double
veneer_inexact_extended
veneer_inexact_integer
veneer_inexact_packed
veneer_inexact_xpacked

        ADR     r0, E_FP_INX
        B       GenerateError

; The "core faulty" routine
; =========================

veneer_corefaulty

        ADR     r0, E_FP_CoreFaulty
        B       GenerateError

; The "core disabled" routine
; ===========================

veneer_coredisabled

        ADRL    r0, E_FP_CoreDisabled
        B       GenerateError

        GBLS    GenerateErrorName
GenerateErrorName SETS            "FP":CC:"$FPCodeType":CC:"_GenerateError"
        IMPORT          $GenerateErrorName

GenerateError
; still in the mode appropriate to an illegal instruction handler
; (SWI26 or UNDEF32).
        ADR     r1, ExceptionRegs
 [ {CONFIG} = 32
        LDR     r11, [r12, #-8]
 |
        LDR     r14, [r12, #r15*4]
        AND     r11, r14, #&fc000003
 ]
        ; retrieve the register values at FPE entry from the current stack
        ; (adjusting the pc value to address the failing instruction rather
        ; than the next one), and record in an argument block for generate
        ; error swi. (clears the FPE stack to empty).
        MOV     r13, r12
        LDMIA   r13!, {r3-r10}
        STMIA   r1!, {r3-r10}
        LDMIA   r13!, {r3-r10}
        SUB     r10, r10, #4
        STMIA   r1!, {r3-r11}
        ADR     r1, ExceptionRegs
        B       $GenerateErrorName

; The global workspace, with its single included context.

FP_Workspace       %           WorkspaceLength

OldVector & 0
BranchTarget & 0
ExceptionRegs % 17*4

        LTORG

        GET h_errors.s
Error_FP_CoreFaulty * Error_FPBase + &80
Error_FP_CoreDisabled * Error_FPBase + &81

        &       -1

        GBLA    Index
Index   SETA    0

ReversingTable
        WHILE   Index < ByteArrayCount
        DCD     ByteArray_$Index, ByteArrayEnd_$Index
Index   SETA    Index+1
        WEND

        &       2                ; 2 indicates reversing tables (above)
        &       FP_Workspace
        &       -1               ; marker for start of error strings

        ErrorBlock FP_INX, "Floating Point Exception : Inexact Result"
        ErrorBlock FP_UFL, "Floating Point Exception : Underflow"
        ErrorBlock FP_DVZ, "Floating Point Exception : Divide By Zero"
        ErrorBlock FP_OFL, "Floating Point Exception : Overflow"
        ErrorBlock FP_IVO, "Floating Point Exception : Invalid Operation"
        ErrorBlock FP_CoreFaulty, "FPE Internal Error : Core Faulty"
        ErrorBlock FP_CoreDisabled, "FPE Internal Error : Core Disabled"

        END
